package simpleex.core;

import java.util.Arrays;
import java.util.Iterator;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

public class MetaDataTest {

  private MetaData metaData;

  @Before
  public void setUp() {
    metaData = new MetaData();
  }

  @Test
  public void testHasTags() {
    Assert.assertTrue(metaData.hasTags());
    Assert.assertFalse(metaData.hasTags("aTag"));

    metaData.addTags("aTag", "bTag");
    Assert.assertTrue(metaData.hasTags());
    Assert.assertTrue(metaData.hasTags("aTag"));
    Assert.assertFalse(metaData.hasTags("aTag", "cTag"));

    metaData.removeTags("aTag");
    Assert.assertTrue(metaData.hasTags());
    Assert.assertFalse(metaData.hasTags("aTag"));
    Assert.assertTrue(metaData.hasTags("bTag"));
  }

  // relies on a certain order
  private void check(final Iterator<String> it, final String... ss) {
    final Iterator<String> it1 = Arrays.asList(ss).iterator();
    while (it1.hasNext() && it.hasNext()) {
      Assert.assertEquals(it1.next(), it.next());
    }
    Assert.assertEquals(it1.hasNext(), it.hasNext());
  }

  // relies on a certain order
  private void checkTags(final String... tags) {
    check(metaData.tags(), tags);
  }

  @Test
  public void testMetaData() {
    Assert.assertTrue(metaData.isEmpty());
    checkTags();
  }

  @Test
  public void testSetTags() {
    metaData.setTags("aTag", "bTag");
    checkTags("aTag", "bTag");
    metaData.setTags("cTag", "bTag");
    checkTags("cTag", "bTag");
    metaData.setTags();
    checkTags();
  }

  @Test
  public void testAddRemoveTags() {
    metaData.addTags("aTag", "bTag");
    checkTags("aTag", "bTag");
    metaData.addTags("cTag");
    checkTags("aTag", "bTag", "cTag");
    metaData.removeTags("bTag");
    checkTags("aTag", "cTag");
    metaData.removeTags("cTag");
    checkTags("aTag");
    metaData.removeTags("aTag");
    Assert.assertTrue(metaData.isEmpty());
  }

  @Test
  public void testHasProperty() {
    Assert.assertFalse(metaData.hasProperty("aProp"));
    metaData.setProperty("aProp", "aValue");
    Assert.assertTrue(metaData.hasProperty("aProp"));
  }

  @Test
  public void testGetSetProperty() {
    metaData.setProperty("aProp", "aValue");
    Assert.assertEquals("aValue", metaData.getProperty("aProp"));
    metaData.setProperty("bProp", "bValue");
    Assert.assertEquals("bValue", metaData.getProperty("bProp"));
    metaData.setProperty("aProp", "anotherValue");
    Assert.assertEquals("anotherValue", metaData.getProperty("aProp"));

    Assert.assertEquals(-1, metaData.getIntegerProperty("iProp", -1));
    metaData.setProperty("iProp", "notAnInteger");
    Assert.assertEquals(-1, metaData.getIntegerProperty("dProp", -1));
    metaData.setIntegerProperty("iProp", 42);
    Assert.assertEquals(42, metaData.getIntegerProperty("iProp", -1));

    Assert.assertEquals(-1.0, metaData.getDoubleProperty("dProp", -1.0), 0.0);
    metaData.setProperty("dProp", "notADouble");
    Assert.assertEquals(-1.0, metaData.getDoubleProperty("dProp", -1.0), 0.0);
    metaData.setDoubleProperty("dProp", 42.0);
    Assert.assertEquals(42.0, metaData.getDoubleProperty("dProp", -1.0), 0.0);

    Assert.assertEquals(false, metaData.getBooleanProperty("bProp", false));
    metaData.setProperty("bProp", "notABoolean");
    Assert.assertEquals(false, metaData.getBooleanProperty("bProp", true));
    metaData.setBooleanProperty("bProp", true);
    Assert.assertEquals(true, metaData.getBooleanProperty("bProp", false));
  }

  @Test
  public void testSetRemoveProperty() {
    metaData.setProperty("aProp", "aValue");
    Assert.assertTrue(metaData.hasProperty("aProp"));
    metaData.removeProperty("aProp");
    Assert.assertFalse(metaData.hasProperty("aProp"));
    Assert.assertTrue(metaData.isEmpty());
  }

  // relies on a certain order
  private void checkPropertyNames(final String... propertyNames) {
    check(metaData.propertyNames(), propertyNames);
  }

  @Test
  public void testPropertyNames() {
    checkPropertyNames();
    metaData.setProperty("aProp", "aValue");
    checkPropertyNames("aProp");
    metaData.removeProperty("aProp");
    checkPropertyNames();
  }
}
